// g_phys.c

#include "g_local.h"

/*

pushmove objects do not obey gravity, and do not interact with each other or trigger fields, but block normal movement and push normal objects when they move.

onground is set for toss objects when they come to a complete rest.  it is set for steping or walking objects 

doors, plats, etc are SOLID_BSP, and MOVETYPE_PUSH
bonus items are SOLID_TRIGGER touch, and MOVETYPE_TOSS
corpses are SOLID_NOT and MOVETYPE_TOSS
crates are SOLID_BBOX and MOVETYPE_TOSS
walking monsters are SOLID_BBOX and MOVETYPE_STEP
flying/floating monsters are SOLID_BBOX and MOVETYPE_FLY

solid_edge items only clip against bsp models.

*/

double	g_frametime;

#define	MOVE_EPSILON	0.01

void SV_Physics_Toss (Entity *ent);


Entity *SV_TestEntityPosition (Entity *ent)
{
	trace_t	trace;

	trace = gi.trace (ent->origin, ent->mins, ent->maxs, ent->origin, MOVE_NORMAL, ent);
	
	if (trace.startsolid)
		return world;
		
	return NULL;
}


/*
================
SV_CheckAllEnts
================
*/
void SV_CheckAllEnts (void)
{
	int			e;
	Entity		*check;

// see if any solid entities are inside the final position
	check = g_edicts + 1;
	for (e = 1; e < globals.num_edicts; e++, check++)
	{
		if (!check->inuse)
			continue;
		if (check->movetype == MOVETYPE_PUSH
		|| check->movetype == MOVETYPE_NONE
		|| check->movetype == MOVETYPE_NOCLIP)
			continue;

		if (SV_TestEntityPosition (check))
			Com_Printf ("entity in invalid position\n");
	}
}

/*
================
SV_CheckVelocity
================
*/
void SV_CheckVelocity (Entity *ent)
{
	int		i;
	float	wishspeed;

//
// bound velocity
//
	for (i=0 ; i<3 ; i++)
	{
		if (IS_NAN(ent->velocity[i]))
		{
			Com_DPrintf ("Got a NaN velocity on %s\n", ent->classname);
			ent->velocity[i] = 0;
		}
		if (IS_NAN(ent->origin[i]))
		{
			Com_DPrintf ("Got a NaN origin on %s\n", ent->classname);
			ent->origin[i] = 0;
		}
/*		if (ent->velocity[i] > sv_maxvelocity.value)
			ent->velocity[i] = sv_maxvelocity.value;
		else if (ent->velocity[i] < -sv_maxvelocity.value)
			ent->velocity[i] = -sv_maxvelocity.value;
*/
	}

	// SV_MAXVELOCITY fix by Maddes
	wishspeed = Length(ent->velocity);
	if (wishspeed > g_sv_maxvelocity->value)
	{
		ent->velocity *= (g_sv_maxvelocity->value / wishspeed);
		wishspeed = g_sv_maxvelocity->value;
	}
}

/*
=============
G_RunThink

Runs thinking code for this frame if necessary
Returns false if the entity removed itself.
=============
*/
qboolean G_RunThink (Entity *ent)
{
	float	thinktime;

	thinktime = ent->nextthink;
	if (thinktime <= 0)
		return true;
	if (thinktime > g_time + 0.001)
		return true;
	
	ent->nextthink = 0;
	if (!ent->think)
		gi.error ("NULL ent->think");
	ent->think (ent);

	if (!ent->inuse)
		return false;

	return true;
}


/*
==================
SV_Impact

Two entities have touched, so run their touch functions
==================
*/
void SV_Impact (Entity *e1, Entity *e2)
{
	if (e1->touch && e1->solid != SOLID_NOT)
		e1->touch (e1, e2);

	if (e2->touch && e2->solid != SOLID_NOT)
		e2->touch (e2, e1);
}


/*
==================
ClipVelocity

Slide off of the impacting object
==================
*/
#define	STOP_EPSILON	0.1

void ClipVelocity (Vec3& in, Vec3& normal, Vec3& out, float overbounce)
{
	float backoff = (in * normal) * overbounce;

	for (int i = 0; i < 3; i++)
	{
		float change = normal[i]*backoff;
		out[i] = in[i] - change;
		if (out[i] > -STOP_EPSILON && out[i] < STOP_EPSILON)
			out[i] = 0;
	}
}


/*
============
SV_FlyMove

The basic solid body movement clip that slides along multiple planes
Returns the clipflags if the velocity was modified (hit something solid)
1 = floor
2 = wall / step
4 = dead stop
If steptrace is not NULL, the trace of any vertical wall hit will be stored
============
*/
#define	MAX_CLIP_PLANES	5
int SV_FlyMove (Entity *ent, float time, trace_t *steptrace)
{
	int			numplanes;
	Vec3		planes[MAX_CLIP_PLANES];
	Vec3		primal_velocity, original_velocity, new_velocity;
	int			i, j;
	trace_t		trace;
	Vec3		end;
	int			blocked;
	
	int numbumps = 4;
	
	blocked = 0;
	original_velocity = ent->velocity;
	primal_velocity = ent->velocity;
	numplanes = 0;
	
	float time_left = time;

	for (int bumpcount = 0; bumpcount < numbumps; bumpcount++)
	{
		end = ent->origin + time_left * ent->velocity;

		trace = gi.trace (ent->origin, ent->mins, ent->maxs, end, MOVE_NORMAL, ent);

		if (trace.startsolid)
		{	// entity is trapped in another solid
			ent->velocity.Clear();
			return 3;
		}

		if (trace.fraction > 0)
		{	// actually covered some distance
			ent->origin = trace.endpos;
			original_velocity = ent->velocity;
			numplanes = 0;
		}

		if (trace.fraction == 1)
			 break;		// moved the entire distance

		if (!trace.ent)
			gi.error ("SV_FlyMove: !trace.ent");

		if (trace.plane.normal[2] > 0.7)
		{
			blocked |= 1;		// floor
			if (trace.ent->solid == SOLID_BSP)
			{
				ent->groundentity = (Entity *)trace.ent;
			}
		}
		if (!trace.plane.normal[2])
		{
			blocked |= 2;		// step
			if (steptrace)
				*steptrace = trace;	// save for player extrafriction
		}

//
// run the impact function
//
		SV_Impact (ent, (Entity *)trace.ent);
		if (!ent->inuse)
			break;		// removed by the impact function

		
		time_left -= time_left * trace.fraction;
		
	// cliped to another plane
		if (numplanes >= MAX_CLIP_PLANES)
		{	// this shouldn't really happen
			ent->velocity.Clear();
			return 3;
		}

		planes[numplanes] = trace.plane.normal;
		numplanes++;

//
// modify original_velocity so it parallels all of the clip planes
//
		for (i=0 ; i<numplanes ; i++)
		{
			ClipVelocity (original_velocity, planes[i], new_velocity, 1);
			for (j=0 ; j<numplanes ; j++)
				if (j != i)
				{
					if (new_velocity * planes[j] < 0)
						break;	// not ok
				}
			if (j == numplanes)
				break;
		}
		
		if (i != numplanes)
		{	// go along this plane
			ent->velocity = new_velocity;
		}
		else
		{	// go along the crease
			if (numplanes != 2)
			{
//				Com_Printf ("clip velocity, numplanes == %i\n",numplanes);
				ent->velocity.Clear();
				return 7;
			}
			Vec3 dir = CrossProduct (planes[0], planes[1]);
			ent->velocity = (ent->velocity * dir) * dir;
		}

//
// if original velocity is against the original velocity, stop dead
// to avoid tiny occilations in sloping corners
//
		if (DotProduct (ent->velocity, primal_velocity) <= 0)
		{
			ent->velocity.Clear();
			return blocked;
		}
	}

	return blocked;
}


void SV_AddGravity (Entity *ent, float scale)
{
	ent->velocity[2] -= scale * g_sv_gravity->value * g_frametime;
}

/*
===============================================================================

PUSHMOVE

===============================================================================
*/

/*
============
SV_PushEntity

Does not change the entity's velocity at all
============
*/
trace_t SV_PushEntity (Entity *ent, Vec3& push)
{
	trace_t	trace;
	
	Vec3 end = ent->origin + push;

	if (ent->solid == SOLID_TRIGGER || ent->solid == SOLID_NOT)
	// only clip against bmodels
		trace = gi.trace (ent->origin, ent->mins, ent->maxs, end, MOVE_NOMONSTERS, ent);
	else
		trace = gi.trace (ent->origin, ent->mins, ent->maxs, end, MOVE_NORMAL, ent);

	ent->origin = trace.endpos;
	G_LinkEntity (ent);
	G_TouchTriggers (ent);

	if (trace.ent)
		SV_Impact (ent, (Entity *)trace.ent);

	return trace;
}					


/*
============
SV_Push

============
*/
qboolean SV_Push (Entity *pusher, Vec3& move)
{
	int			i, e;
	Entity		*check, *block;
	Vec3		mins, maxs;
	Vec3		pushorig;
	int			num_moved;
	Entity		*moved_edict[MAX_EDICTS];
	Vec3		moved_from[MAX_EDICTS];

	mins = pusher->absmin + move;
	maxs = pusher->absmax + move;

	pushorig = pusher->origin;
	
// move the pusher to its final position

	pusher->origin += move;
	G_LinkEntity (pusher);

// see if any solid entities are inside the final position
	num_moved = 0;
	check = g_edicts + 1;
	for (e = 1; e < globals.num_edicts; e++, check++)
	{
		if (!check->inuse)
			continue;
		if (check->movetype == MOVETYPE_PUSH
		|| check->movetype == MOVETYPE_NONE
		|| check->movetype == MOVETYPE_NOCLIP)
			continue;

		pusher->solid = SOLID_NOT;
		block = SV_TestEntityPosition (check);
		pusher->solid = SOLID_BSP;
		if (block)
			continue;

	// if the entity is standing on the pusher, it will definitely be moved
		if (check->groundentity != pusher)
		{
			if ( check->absmin[0] >= maxs[0]
			|| check->absmin[1] >= maxs[1]
			|| check->absmin[2] >= maxs[2]
			|| check->absmax[0] <= mins[0]
			|| check->absmax[1] <= mins[1]
			|| check->absmax[2] <= mins[2] )
				continue;

		// see if the ent's bbox is inside the pusher's final position
			if (!SV_TestEntityPosition (check))
				continue;
		}

		moved_from[num_moved] = check->origin;
		moved_edict[num_moved] = check;
		num_moved++;

		// try moving the contacted entity 
		check->origin += move;
		block = SV_TestEntityPosition (check);
		if (!block)
		{	// pushed ok
			G_LinkEntity (check);
			continue;
		}

		// if it is ok to leave in the old position, do it
		check->origin -= move;
		block = SV_TestEntityPosition (check);
		if (!block)
		{
			num_moved--;
			continue;
		}

	// if it is still inside the pusher, block
		if (check->mins[0] == check->maxs[0])
		{
			G_LinkEntity (check);
			continue;
		}
		if (check->solid == SOLID_NOT || check->solid == SOLID_TRIGGER)
		{	// corpse
			check->mins[0] = check->mins[1] = 0;
			check->maxs = check->mins;
			G_LinkEntity (check);
			continue;
		}
		
		pusher->origin = pushorig;
		G_LinkEntity (pusher);

		// if the pusher has a "blocked" function, call it
		// otherwise, just stay in place until the obstacle is gone
		if (pusher->blocked)
			pusher->blocked (pusher, check);
		
	// move back any entities we already moved
		for (i=0 ; i<num_moved ; i++)
		{
			moved_edict[i]->origin = moved_from[i];
			G_LinkEntity (moved_edict[i]);
		}
		return false;
	}

	return true;
}


void SV_PushMove (Entity *pusher, float movetime)
{
	if (!pusher->velocity[0] && !pusher->velocity[1] && !pusher->velocity[2])
	{
		pusher->ltime += movetime;
		return;
	}

	Vec3 move = pusher->velocity * movetime;

	if ( SV_Push(pusher, move) )
		pusher->ltime += movetime;
}

void SV_Physics_Pusher (Entity *ent)
{
	float	thinktime;
	float	oldltime;
	float	movetime;
Vec3 oldorg, move;
float	l;

	oldltime = ent->ltime;
	
	thinktime = ent->nextthink;
	if (thinktime < ent->ltime + g_frametime)
	{
		movetime = thinktime - ent->ltime;
		if (movetime < 0)
			movetime = 0;
	}
	else
		movetime = g_frametime;

	if (movetime)
	{
		SV_PushMove (ent, movetime);	// advances ent->ltime if not blocked
	}
		
	if (thinktime > oldltime && thinktime <= ent->ltime)
	{
VectorCopy (ent->origin, oldorg);
		ent->nextthink = 0;
		if (ent->think)
			ent->think (ent);
		if (!ent->inuse)
			return;
VectorSubtract (ent->origin, oldorg, move);

l = VectorLength(move);
if (l > 1.0/64)
{
//	Com_Printf ("**** snap: %f\n", Length (l));
	VectorCopy (oldorg, ent->origin);
	SV_Push (ent, move);
}

	}

}


/*
=============
SV_Physics_None

Non moving objects can only think
=============
*/
void SV_Physics_None (Entity *ent)
{
// regular thinking
	G_RunThink (ent);
}

/*
=============
SV_Physics_Noclip

A moving object that doesn't obey physics
=============
*/
void SV_Physics_Noclip (Entity *ent)
{
// regular thinking
	if (!G_RunThink (ent))
		return;

	ent->angles += ent->avelocity * g_frametime;
//	ent->angles += ent->avelocity * g_frametime;
	ent->origin += ent->velocity * g_frametime;

	G_LinkEntity (ent);
}

/*
==============================================================================

TOSS / BOUNCE

==============================================================================
*/

/*
=============
SV_CheckWaterTransition

=============
*/
void SV_CheckWaterTransition (Entity *ent)
{
	int		cont;

	cont = gi.pointcontents (ent->origin);
	if (!ent->watertype)
	{	// just spawned here
		ent->watertype = cont;
		ent->waterlevel = 1;
		return;
	}
	
	if (cont <= CONTENTS_WATER)
	{
		if (ent->watertype == CONTENTS_EMPTY)
		{	// just crossed into water
			G_Sound (ent, 0, "misc/h2ohit1.wav");
		}		
		ent->watertype = cont;
		ent->waterlevel = 1;
	}
	else
	{
		if (ent->watertype != CONTENTS_EMPTY)
		{	// just crossed into water
			G_Sound (ent, 0, "misc/h2ohit1.wav");
		}		
		ent->watertype = CONTENTS_EMPTY;
		ent->waterlevel = cont;
	}
}

/*
=============
SV_Physics_Toss

Toss, bounce, and fly movement.  When onground, do nothing.
=============
*/
void SV_Physics_Toss (Entity *ent)
{
	trace_t	trace;
	Vec3	move;
	float	backoff;

// regular thinking
	if (!G_RunThink (ent))
		return;

	if (ent->velocity[2] > 0)
		ent->groundentity = NULL;

// if onground, return without moving
	if (ent->groundentity)
		return;

	SV_CheckVelocity (ent);

// add gravity
	if (ent->movetype != MOVETYPE_FLY)
		SV_AddGravity (ent, 1.0);

// move angles
	ent->angles += ent->avelocity * g_frametime;

// move origin
	move = ent->velocity * g_frametime;
	trace = SV_PushEntity (ent, move);
	if (trace.fraction == 1)
		return;
	if (!ent->inuse)
		return;
	
	if (ent->movetype == MOVETYPE_BOUNCE)
		backoff = 1.5;
	else
		backoff = 1;

	ClipVelocity (ent->velocity, trace.plane.normal, ent->velocity, backoff);

// stop if on ground
	if (trace.plane.normal[2] > 0.7)
	{		
		if (ent->velocity[2] < 60 || ent->movetype != MOVETYPE_BOUNCE )
		{
			ent->groundentity = (Entity *)trace.ent;
			ent->velocity.Clear();
			ent->avelocity.Clear();
		}
	}
	
// check for in water
	SV_CheckWaterTransition (ent);
}

/*
===============================================================================

STEPPING MOVEMENT

===============================================================================
*/

/*
=============
SV_Physics_Step

Monsters freefall when they don't have a ground entity, otherwise
all movement is done with discrete steps.

This is also used for objects that have become still on the ground, but
will fall if the floor is pulled out from under them.
FIXME: is this true?
=============
*/
void SV_Physics_Step (Entity *ent)
{
	qboolean	hitsound;

// freefall if not onground
	if (!ent->groundentity && !(ent->flags & (FL_FLY|FL_SWIM)))
	{
		if (ent->velocity[2] < g_sv_gravity->value * -0.1)
			hitsound = true;
		else
			hitsound = false;

		SV_AddGravity (ent, 1.0);
		SV_CheckVelocity (ent);
		SV_FlyMove (ent, g_frametime, NULL);
		G_LinkEntity (ent);
		G_TouchTriggers (ent);

		if (ent->groundentity)	// just hit ground
		{
			if (hitsound)
				G_Sound (ent, 0, "demon/dland2.wav");
		}
	}

// regular thinking
	G_RunThink (ent);
	
	SV_CheckWaterTransition (ent);
}


/*
================
G_RunEntity
================
*/
void G_RunEntity (Entity *ent)
{
	if (ent->lastruntime == g_time)
		return;
	ent->lastruntime = g_time;

	ent->old_origin = ent->origin;

	switch (ent->movetype)
	{
	case MOVETYPE_NONE:
		SV_Physics_None (ent);
		break;
	case MOVETYPE_PUSH:
		SV_Physics_Pusher (ent);
		break;
	case MOVETYPE_NOCLIP:
		SV_Physics_Noclip (ent);
		break;
	case MOVETYPE_STEP:
		SV_Physics_Step (ent);
		break;
	case MOVETYPE_TOSS:
	case MOVETYPE_BOUNCE:
	case MOVETYPE_FLY:
		SV_Physics_Toss (ent);
		break;
	default:
		gi.error ("G_RunEntity: bad movetype %i", ent->movetype);
	}
}


/*
================
G_RunNewmis
================
*/
void G_RunNewmis (Entity *ent, double time)
{
	double save_frametime = g_frametime;
	g_frametime = time;
	G_RunEntity (ent);
	g_frametime = save_frametime;
}


void G_UpdateClientStats ()
{
	Entity *ent;
	int		*stats;

	for (int i = 0; i < MAX_CLIENTS; i++) {
		ent = &g_edicts[i];
		if (!ent->inuse)
			continue;

		stats = (int *)ent->stats;
		assert (stats);

//		memset (stats, 0, sizeof(stats_t));
		ent->weaponmodelindex = gi.modelindex(ent->weaponmodel); // FIXME: we don't wanna do this every frame!
		stats[STAT_HEALTH] = ent->health;
		stats[STAT_AMMO] = ent->currentammo;
		stats[STAT_ARMOR] = ent->armorvalue;
		stats[STAT_SHELLS] = ent->ammo_shells;
		stats[STAT_NAILS] = ent->ammo_nails;
		stats[STAT_ROCKETS] = ent->ammo_rockets;
		stats[STAT_CELLS] = ent->ammo_cells;
		if (!ent->spectator)
			stats[STAT_ACTIVEWEAPON] = ent->weapon;
		else
			stats[STAT_ACTIVEWEAPON] = 0;
		stats[STAT_ITEMS] = (int)ent->items;
	}
}

int force_retouch;

/*
================
G_RunFrame
================
*/
void G_RunFrame (double time)
{
	int		i;
	Entity	*ent;

	g_frametime = time;
	g_time += time;

//
// treat each object in turn
// even the world gets a chance to think
//
	ent = g_edicts;
	for (i=0 ; i<globals.num_edicts ; i++, ent++)
	{
		if (!ent->inuse)
			continue;

		if (force_retouch)
		{
			// force retouch even for stationary
			G_LinkEntity (ent);		// FIXME
			G_TouchTriggers (ent);
		}

		if (i > 0 && i <= MAX_CLIENTS)
			continue;		// clients are run directly from packets

		G_RunEntity (ent);
	}
	
	if (force_retouch)
		force_retouch--;

	G_UpdateClientStats ();
}
